<?php

declare(strict_types=1);

namespace Drupal\form_alterer\Services;

use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Cache\CacheBackendInterface;
use Drupal\Core\Plugin\DefaultPluginManager;
use Drupal\form_alterer\Annotation\FormAlterer;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\Plugin\Discovery\YamlDiscoveryDecorator;
use Drupal\Core\Plugin\Discovery\AnnotatedClassDiscovery;
use Drupal\form_alterer\Plugin\FormAlterer\FormAltererInterface;
use Drupal\Core\Plugin\Discovery\ContainerDerivativeDiscoveryDecorator;
use Drupal\Component\Annotation\Plugin\Discovery\AnnotationBridgeDecorator;
use Drupal\form_alterer\Plugin\FormAlterer\FormAltererDefault;

/**
 * FormAlterer plugin manager.
 */
final class FormAltererPluginManager extends DefaultPluginManager implements FormAltererPluginManagerInterface {

  /**
   * {@inheritDoc}
   */
  protected $defaults = [
    'form_ids' => [],
    'static_overrides' => [],
    'specific_static_overrides' => [],
    'execution_order' => 1,
    'category' => 'default',
  ];

  /**
   * The form alterer prototypes.
   *
   * @var \Drupal\form_alterer\Plugin\FormAlterer\FormAltererInterface[]
   */
  protected $prototypes = [];

  /**
   * Constructs the object.
   */
  public function __construct(\Traversable $namespaces, CacheBackendInterface $cache_backend, ModuleHandlerInterface $module_handler) {
    parent::__construct('Plugin/FormAlterer', $namespaces, $module_handler, FormAltererInterface::class, FormAlterer::class);
    $this->alterInfo('form_alterer_info');
    $this->setCacheBackend($cache_backend, 'form_alterer_plugins');
  }

  /**
   * {@inheritdoc}
   */
  public function processDefinition(&$definition, $plugin_id) {
    parent::processDefinition($definition, $plugin_id);
    if (!isset($definition['class'])) {
      $definition['class'] = FormAltererDefault::class;
    }
  }

  /**
   * {@inheritdoc}
   */
  protected function getDiscovery() {
    if (!$this->discovery) {
      $discovery = new AnnotatedClassDiscovery($this->subdir, $this->namespaces, $this->pluginDefinitionAnnotationName, $this->additionalAnnotationNamespaces);
      $discovery = new YamlDiscoveryDecorator($discovery, 'form_alterers', $this->moduleHandler->getModuleDirectories());
      $discovery
        ->addTranslatableProperty('label')
        ->addTranslatableProperty('description')
        ->addTranslatableProperty('category');
      $discovery = new AnnotationBridgeDecorator($discovery, $this->pluginDefinitionAnnotationName);
      $discovery = new ContainerDerivativeDiscoveryDecorator($discovery);
      $this->discovery = $discovery;
    }
    return $this->discovery;
  }

  /**
   * Returns plugin instances based on the form id given.
   */
  public function getInstancesFromFormId($form_id): array {
    $definitions = array_filter($this->getDefinitions(), function ($definition) use ($form_id) {
      return in_array($form_id, $definition['form_ids']);
    });
    $instances = [];
    foreach ($definitions as $id => $definition) {
      if (!isset($this->prototypes[$id])) {
        $this->prototypes[$id] = $this->createInstance($id);
      }
      $instances[$id] = $this->prototypes[$id];
    }
    return $instances;
  }

  /**
   * Alters the given forms with form alterer plugins.
   */
  public function alterForm(&$form, FormStateInterface $form_state, $form_id) {
    $base_form_id = $form_state->getBuildInfo()['base_form_id'] ?? NULL;
    $form_instances = [];
    // Check for instances with a base form id.
    if ($base_form_id) {
      foreach ($this->getInstancesFromFormId($base_form_id) as $id => $instance) {
        if (!isset($form_instances[$id])) {
          $form_instances[$id] = $instance;
        }
      }
    }
    // Check for instances with a normal form id.
    foreach ($this->getInstancesFromFormId($form_id) as $id => $instance) {
      if (!isset($form_instances[$id])) {
        $form_instances[$id] = $instance;
      }
    }
    $executable_instances = [];
    // Sort instances by execution order.
    foreach ($form_instances as $id => $instance) {
      $definition = $instance->getPluginDefinition();
      $execution_order = $definition['execution_order'];
      $executable_instances[$execution_order][$id] = $instance;
    }
    // Execute instances by execution order.
    foreach ($executable_instances as $list) {
      foreach ($list as $instance) {
        $instance->alterForm($form, $form_state, $form_id);
      }
    }
    // Add the instances to the form state.
    $form_state->set('form_alterer_plugins', $form_instances);
  }

}
